#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# CheckboxNG documentation build configuration file, created by
# sphinx-quickstart on Wed Feb 13 11:18:39 2013.
#
# This file is execfile()d with the current directory set to its containing
# dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import sys

try:
    import plainbox
except ImportError as exc:
    raise SystemExit("plainbox has to be importable")
else:
    # Inject mock modules so that we can build the
    # documentation without having the real stuff available
    from plainbox.vendor import mock
    for mod_name in ['lxml', 'xlsxwriter', 'requests', 'requests.exceptions',
                     'dbus', 'dbus.lowlevel', 'dbus.exceptions',
                     'dbus._compat', 'dbus.service', '_dbus_bindings']:
        sys.modules[mod_name] = mock.Mock()
        print("Mocked {}".format(mod_name))
try:
    import checkbox_ng
except ImportError:
    raise SystemExit("checkbox_ng has to be importable")

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath('.'))

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.doctest', 'sphinx.ext.todo',
              'sphinx.ext.coverage', 'sphinx.ext.viewcode',
              'sphinx.ext.intersphinx']


def check_object_path(key, url, path):
    if os.path.isfile(path):
        return {key: (url, path)}
    else:
        return {key: (url, None)}


intersphinx_mapping = {}
intersphinx_mapping.update(
    check_object_path(
        'python', 'http:/docs.python.org/',
        '/usr/share/doc/python{}/html/objects.inv'.format(
            '.'.join([str(x) for x in sys.version_info[0:2]]))))
intersphinx_mapping.update(
    check_object_path(
        'plainbox', 'http://plainbox.readthedocs.org/en/latest/',
        '/usr/share/doc/python3-plainbox-doc/html/objects.inv'))

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'CheckboxNG'
copyright = '2013-2014, Zygmunt Krynicki'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = "{0[0]}.{0[1]}".format(checkbox_ng.__version__)
# The full version, including alpha/beta/rc tags.
release = "{0[0]}.{0[1]}.{0[2]}.{0[3]}.{0[4]}".format(checkbox_ng.__version__)

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# The reST default role (used for this markup: `text`) to use for all
# documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output -------------------------------------------------

html_theme_path = ['_templates']
html_theme = 'ubuntu1204'
html_theme_options = {}

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'CheckboxNGdoc'


# -- Options for LaTeX output ------------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
# documentclass [howto/manual]).
latex_documents = [
    ('index', 'CheckboxNG.tex', 'CheckboxNG Documentation',
     'Zygmunt Krynicki', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'checkbox_ng', 'CheckboxNG Documentation',
     ['Zygmunt Krynicki'], 1),
    ('stack', 'checkbox-stack', 'description of the checkbox stack',
     ['Zygmunt Krynicki'], 1),
]

# If true, show URL addresses after external links.
man_show_urls = True


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'CheckboxNG', 'CheckboxNG Documentation',
     'Zygmunt Krynicki', 'CheckboxNG', 'One line description of project.',
     'Miscellaneous'),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'



# This file is part of Checkbox.
#
# Copyright 2013 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox_ng.config` -- CheckBoxNG configuration
=====================================================
"""

from gettext import gettext as _
import itertools
import os

from plainbox.impl.applogic import PlainBoxConfig
from plainbox.impl.secure import config


SECURE_ID_PATTERN = r"^[a-zA-Z0-9]{15}$|^[a-zA-Z0-9]{18}$"


class CheckBoxConfig(PlainBoxConfig):
    """
    Configuration for checkbox-ng
    """

    secure_id = config.Variable(
        section="sru",
        help_text=_("Secure ID of the system"),
        validator_list=[config.PatternValidator(SECURE_ID_PATTERN)])

    submit_to_c3 = config.Variable(
        section="submission",
        help_text=_("Whether to send the submission data to c3"))

    submit_to_hexr = config.Variable(
        section="submission",
        help_text=_("Whether to also send the submission data to HEXR"),
        kind=bool)

    # TODO: Add a validator to check if email looks fine
    email_address = config.Variable(
        section="sru",
        help_text=_("Email address to log into the Launchpad HWDB"))

    # TODO: Add a validator to check if URL looks fine
    c3_url = config.Variable(
        section="sru",
        help_text=_("URL of the certification website"),
        default="https://certification.canonical.com/submissions/submit/")

    # TODO: Add a validator to check if URL looks fine
    lp_url = config.Variable(
        section="sru",
        help_text=_("URL of the launchpad hardware database"),
        default="https://launchpad.net/+hwdb/+submit")

    fallback_file = config.Variable(
        section="sru",
        help_text=_("Location of the fallback file"))

    whitelist = config.Variable(
        section="sru",
        help_text=_("Optional whitelist with which to run SRU testing"))

    class Meta(PlainBoxConfig.Meta):
        # TODO: properly depend on xdg and use real code that also handles
        # XDG_CONFIG_HOME.
        #
        # NOTE: filename_list is composed of checkbox and plainbox variables,
        # mixed so that:
        # - checkbox takes precedence over plainbox
        # - ~/.config takes precedence over /etc
        filename_list = list(
            itertools.chain(
                *zip(
                    PlainBoxConfig.Meta.filename_list, (
                        '/etc/xdg/checkbox.conf',
                        os.path.expanduser('~/.config/checkbox.conf')))))



# This file is part of Checkbox.
#
# Copyright 2012-2014 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox_ng.main` -- entry points for command line tools
==============================================================
"""

import logging

from plainbox.impl.logging import setup_logging

from checkbox_ng.tools import CheckboxLauncherTool
from checkbox_ng.tools import CheckboxServiceTool
from checkbox_ng.tools import CheckboxSubmitTool
from checkbox_ng.tools import CheckboxTool


logger = logging.getLogger("checkbox.ng.main")


def main(argv=None):
    """
    checkbox command line utility
    """
    raise SystemExit(CheckboxTool().main(argv))


def service(argv=None):
    """
    checkbox-service command line utility
    """
    raise SystemExit(CheckboxServiceTool().main(argv))


def submit(argv=None):
    """
    checkbox-submit command line utility
    """
    raise SystemExit(CheckboxSubmitTool().main(argv))


def launcher(argv=None):
    """
    checkbox-launcher command line utility
    """
    raise SystemExit(CheckboxLauncherTool().main(argv))


# Setup logging before anything else starts working.
# If we do it in main() or some other place then unit tests will see
# "leaked" log files which are really closed when the runtime shuts
# down but not when the tests are finishing
setup_logging()



# This file is part of Checkbox.
#
# Copyright 2013 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox_ng.dbus_ex` -- DBus Extensions
=============================================
"""

__all__ = [
    'service',
    'exceptions',
    'Signature',
    'Struct',
    'types',
    'INTROSPECTABLE_IFACE',
    'PEER_IFACE',
    'PROPERTIES_IFACE',
    'OBJECT_MANAGER_IFACE',
]

import re
import os

if os.getenv("MOCK_DBUS") == "yes":
    import sys
    from plainbox.vendor import mock
    for name in (
        '_dbus_bindings',
        'dbus',
        'dbus._compat',
        'dbus.exceptions',
        'dbus.lowlevel',
        'dbus.mainloop',
        'dbus.mainloop.glib',
        'dbus.service',
        'gi',
        'gi.repository',
    ):
        sys.modules[name] = mock.MagicMock(name=name)
    sys.modules['dbus'].service.Object = object
    sys.modules['dbus'].service.Interface = object
    sys.modules['dbus'].service.InterfaceType = type


from dbus import INTROSPECTABLE_IFACE
from dbus import PEER_IFACE
from dbus import PROPERTIES_IFACE
from dbus import Signature
from dbus import Struct
from dbus import exceptions
from dbus import types

OBJECT_MANAGER_IFACE = "org.freedesktop.DBus.ObjectManager"

from checkbox_ng.dbus_ex import service


def mangle_object_path(path):
    """
    "Mangles" the provided candidate dbus object path to ensure it complies
    with the dbus specification. Returns the mangled path.
    """
    # TODO: It just enforces the valid characters rule, not the rest of the
    # DBus path construction rules
    return re.sub(r"[^a-zA-Z0-9_/]", "_", path)



# This file is part of Checkbox.
#
# Copyright 2013, 2014 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox_ng.service` -- DBus service for CheckBox
=======================================================
"""

from gettext import gettext as _
import collections
import functools
import itertools
import logging
import sys

try:
    from inspect import Signature
except ImportError:
    try:
        from plainbox.vendor.funcsigs import Signature
    except ImportError:
        raise SystemExit(_("DBus parts require 'funcsigs' from pypi."))
from plainbox.abc import IJobResult
from plainbox.impl.job import JobDefinition
from plainbox.impl.result import JobResultBuilder
from plainbox.impl.secure.qualifiers import select_jobs
from plainbox.impl.session import JobState
from plainbox.vendor import extcmd
from plainbox.vendor.morris import remove_signals_listeners

from checkbox_ng import dbus_ex as dbus
from checkbox_ng.dbus_ex import OBJECT_MANAGER_IFACE
from checkbox_ng.dbus_ex import mangle_object_path

logger = logging.getLogger("checkbox.ng.service")

_BASE_IFACE = "com.canonical.certification."

SERVICE_IFACE = _BASE_IFACE + "PlainBox.Service1"
SESSION_IFACE = _BASE_IFACE + "PlainBox.Session1"
PROVIDER_IFACE = _BASE_IFACE + "PlainBox.Provider1"
JOB_IFACE = _BASE_IFACE + "PlainBox.JobDefinition1"
JOB_RESULT_IFACE = _BASE_IFACE + "PlainBox.Result1"
JOB_STATE_IFACE = _BASE_IFACE + "PlainBox.JobState1"
WHITELIST_IFACE = _BASE_IFACE + "PlainBox.WhiteList1"
CHECKBOX_JOB_IFACE = _BASE_IFACE + "CheckBox.JobDefinition1"
RUNNING_JOB_IFACE = _BASE_IFACE + "PlainBox.RunningJob1"


class PlainBoxObjectWrapper(dbus.service.ObjectWrapper):
    """
    Wrapper for exporting PlainBox object over DBus.

    Allows to keep the python object logic separate from the DBus counterpart.
    Has a set of utility methods to publish the object and any children objects
    to DBus.
    """

    # Use a different logger for the translate decorator.
    # This is just so that we don't spam people that want to peek
    # at the service module.
    _logger = logging.getLogger("plainbox.dbus.service.translate")

    def __init__(self,
                 native,
                 conn=None, object_path=None, bus_name=None,
                 **kwargs):
        super(PlainBoxObjectWrapper, self).__init__(
            native, conn, object_path, bus_name)
        logger.debug(
            _("Created DBus wrapper %s for: %r"), id(self), self.native)
        self.__shared_initialize__(**kwargs)

    def __del__(self):
        logger.debug(_("DBus wrapper %s died"), id(self))

    def __shared_initialize__(self, **kwargs):
        """
        Optional initialize method that can use any unused keyword arguments
        that were originally passed to __init__(). This makes it far easier to
        subclass as __init__() is rather complicated.

        Inspired by STANDARD GENERIC FUNCTION SHARED-INITIALIZE
        See hyperspec page: http://clhs.lisp.se/Body/f_shared.htm
        """

    def _get_preferred_object_path(self):
        """
        Return the preferred object path of this object on DBus
        """
        return "/plainbox/{}/{}".format(
            self.native.__class__.__name__, id(self.native))

    def publish_self(self, connection):
        """
        Publish this object to the connection
        """
        # Don't publish this object if it's already on the required connection
        # TODO: check if we can just drop this test and publish unconditionally
        try:
            if self.connection is connection:
                return
        except AttributeError:
            pass
        object_path = self._get_preferred_object_path()
        self.add_to_connection(connection, object_path)
        logger.debug(_("Published DBus wrapper for %r as %s"),
                     self.native, object_path)

    def publish_related_objects(self, connection):
        """
        Publish this and any other objects to the connection

        Do not send ObjectManager events, just register any additional objects
        on the bus. By default only the object itself is published but
        collection managers are expected to publish all of the children here.

        This method is meant to be called only once, soon after the object
        is constructed but before :meth:`publish_managed_objects()` is called.
        """
        self.publish_self(connection)

    def publish_managed_objects(self):
        """
        This method is specific to ObjectManager, it basically adds children
        and sends the right events. This is a separate stage so that the whole
        hierarchy can first put all of the objects on the bus and then tell the
        world about it in one big signal message.

        This method is meant to be called only once, soon after
        :meth:`publish_related_objects()` was called.
        """

    @classmethod
    def translate(cls, func):
        """
        Decorator for Wrapper methods.

        The decorated method does not need to manually lookup objects when the
        caller (across DBus) passes an object path. Type information is
        provided using parameter annotations.

        The annotation accepts DBus type expressions (but in practice it is
        very limited). For the moment it cannot infer the argument types from
        the decorator for dbus.service.method.
        """
        if sys.version_info[0:2] >= (3, 5):
            sig = Signature.from_callable(func)
        else:
            sig = Signature.from_function(func)

        def translate_o(object_path):
            try:
                obj = cls.find_object_by_path(object_path)
            except KeyError as exc:
                raise dbus.exceptions.DBusException((
                    _("object path {0} does not designate an existing"
                      " object")).format(exc))
            else:
                return obj.native

        def translate_ao(object_path_list):
            try:
                obj_list = [cls.find_object_by_path(object_path)
                            for object_path in object_path_list]
            except KeyError as exc:
                raise dbus.exceptions.DBusException((
                    _("object path {0} does not designate an existing"
                      " object")).format(exc))
            else:
                return [obj.native for obj in obj_list]

        def translate_return_o(obj):
            if isinstance(obj, PlainBoxObjectWrapper):
                cls._logger.warning(
                    _("Application error: %r should have returned native"
                      " object but returned wrapper instead"), func)
                return obj
            try:
                return cls.find_wrapper_by_native(obj)
            except KeyError:
                raise dbus.exceptions.DBusException(
                    _("({0}) internal error, unable to lookup object"
                      " wrapper").format('o'))

        def translate_return_ao(object_list):
            try:
                return dbus.types.Array([
                    cls.find_wrapper_by_native(obj)
                    for obj in object_list
                ], signature='o')
            except KeyError:
                raise dbus.exceptions.DBusException(
                    _("({0}) internal error, unable to lookup object"
                      " wrapper").format('ao'))

        def translate_return_a_brace_so_brace(mapping):
            try:
                return dbus.types.Dictionary({
                    key: cls.find_wrapper_by_native(value)
                    for key, value in mapping.items()
                }, signature='so')
            except KeyError:
                raise dbus.exceptions.DBusException(
                    _("({0}) internal error, unable to lookup object"
                      " wrapper").format('a{so}'))

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            bound = sig.bind(*args, **kwargs)
            cls._logger.debug(
                _("wrapped %s called with %s"), func, bound.arguments)
            for param in sig.parameters.values():
                if param.annotation is Signature.empty:
                    pass
                elif param.annotation == 'o':
                    object_path = bound.arguments[param.name]
                    bound.arguments[param.name] = translate_o(object_path)
                elif param.annotation == 'ao':
                    object_path_list = bound.arguments[param.name]
                    bound.arguments[param.name] = translate_ao(
                        object_path_list)
                elif param.annotation in ('s', 'as'):
                    strings = bound.arguments[param.name]
                    bound.arguments[param.name] = strings
                else:
                    raise ValueError(
                        _("unsupported translation {!r}").format(
                            param.annotation))
            cls._logger.debug(
                _("unwrapped %s called with %s"), func, bound.arguments)
            retval = func(**bound.arguments)
            cls._logger.debug(_("unwrapped %s returned %r"), func, retval)
            if sig.return_annotation is Signature.empty:
                pass
            elif sig.return_annotation == 'o':
                retval = translate_return_o(retval)
            elif sig.return_annotation == 'ao':
                retval = translate_return_ao(retval)
            elif sig.return_annotation == 'a{so}':
                retval = translate_return_a_brace_so_brace(retval)
            else:
                raise ValueError(
                    _("unsupported translation {!r}").format(
                        sig.return_annotation))
            cls._logger.debug(_("wrapped %s returned  %r"), func, retval)
            return retval
        return wrapper


class JobDefinitionWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing JobDefinition objects on DBus.

    .. note::
        Life cycle of JobDefinition wrappers is associated _either_
        with a Provider wrapper or with a Session wrapper, depending
        on if the job itself is generated or not.
    """

    HIDDEN_INTERFACES = frozenset([
        OBJECT_MANAGER_IFACE,
    ])

    # Some internal helpers

    def __shared_initialize__(self, **kwargs):
        self._checksum = self.native.checksum
        self._is_generated = False

    def _get_preferred_object_path(self):
        # TODO: this clashes with providers, maybe use a random ID instead
        return "/plainbox/job/{}".format(self._checksum)

    # PlainBox properties

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def name(self):
        # XXX: name should be removed but for now it should just return the
        # full id instead of the old name (name may be very well gone)
        return self.native.id

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def summary(self):
        return self.native.summary

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def tr_summary(self):
        return self.native.tr_summary()

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def id(self):
        return self.native.id

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def partial_id(self):
        return self.native.partial_id

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def description(self):
        return self.native.description or ""

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def tr_description(self):
        return self.native.tr_description() or ""

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def checksum(self):
        # This is a bit expensive to compute so let's keep it cached
        return self._checksum

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def requires(self):
        return self.native.requires or ""

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="s")
    def depends(self):
        return self.native.depends or ""

    @dbus.service.property(dbus_interface=JOB_IFACE, signature="d")
    def estimated_duration(self):
        return self.native.estimated_duration or -1

    # PlainBox methods

    @dbus.service.method(dbus_interface=JOB_IFACE,
                         in_signature='', out_signature='as')
    def GetDirectDependencies(self):
        return dbus.Array(
            self.native.get_direct_dependencies(), signature="s")

    @dbus.service.method(dbus_interface=JOB_IFACE,
                         in_signature='', out_signature='as')
    def GetResourceDependencies(self):
        return dbus.Array(
            self.native.get_resource_dependencies(), signature="s")

    @dbus.service.method(dbus_interface=CHECKBOX_JOB_IFACE,
                         in_signature='', out_signature='as')
    def GetEnvironSettings(self):
        return dbus.Array(self.native.get_environ_settings(), signature='s')

    # CheckBox properties

    @dbus.service.property(dbus_interface=CHECKBOX_JOB_IFACE, signature="s")
    def plugin(self):
        return self.native.plugin

    @dbus.service.property(dbus_interface=CHECKBOX_JOB_IFACE, signature="s")
    def via(self):
        # NOTE: the via property is emulated after an incompatible API change
        # that we didn't want to really expose to checkbox-gui. The property
        # no longer belongs to the job definition object. Instead it has
        # to be looked up in the per-session job state object.
        #
        # The downside is that it means that we need to find the job in the
        # state map somewhere and it's not easy to get a hold of that from
        # each job wrapper. So we cheat and look at the (only one possible)
        # session object, conveniently exposed from ServiceWrapper, to find
        # the associated job state object.
        session_obj = ServiceWrapper.get_session_obj()
        if session_obj is None:
            return ""
        try:
            state = session_obj.job_state_map[self.native.id]
        except KeyError:
            # See: LP: #1412660, session_obj can be None and event if it isn't,
            # session_obj.job_state_map might *not* have self.native.id in it.
            return ""
        if state.via_job is None:
            return ""
        return state.via_job.checksum

    @dbus.service.property(
        dbus_interface=CHECKBOX_JOB_IFACE, signature="(suu)")
    def origin(self):
        if self.native.origin is not None:
            return dbus.Struct([
                str(self.native.origin.source),
                self.native.origin.line_start,
                self.native.origin.line_end
            ], signature="suu")
        else:
            return dbus.Struct(["", 0, 0], signature="suu")

    @dbus.service.property(dbus_interface=CHECKBOX_JOB_IFACE, signature="s")
    def command(self):
        return self.native.command or ""

    @dbus.service.property(dbus_interface=CHECKBOX_JOB_IFACE, signature="s")
    def environ(self):
        return self.native.environ or ""

    @dbus.service.property(dbus_interface=CHECKBOX_JOB_IFACE, signature="s")
    def user(self):
        return self.native.user or ""


class TestPlanWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing testplan objects as whitelist on DBus
    """

    HIDDEN_INTERFACES = frozenset([
        OBJECT_MANAGER_IFACE,
    ])

    # Some internal helpers

    def _get_preferred_object_path(self):
        return "/plainbox/whitelist/{}".format(self.native.checksum)

    # Value added

    @dbus.service.property(dbus_interface=WHITELIST_IFACE, signature="s")
    def name(self):
        """
        name of this testplan
        """
        return self.native.name or ""

    @dbus.service.property(dbus_interface=WHITELIST_IFACE, signature="s")
    def partial_id(self):
        """
        partial_id of this testplan
        """
        return self.native.partial_id or ""


class JobResultWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing JobResult objects on DBus.

    This wrapper class exposes two mutable properties, 'outcome'
    and 'comments'. Changing them either through native python APIs
    or through DBus property API will result in synchronized updates
    as well as property change notification signals being sent.
    """

    HIDDEN_INTERFACES = frozenset([
        OBJECT_MANAGER_IFACE,
    ])

    # Value added

    @dbus.service.property(dbus_interface=JOB_RESULT_IFACE, signature="s")
    def outcome(self):
        """
        outcome of the job

        The result is one of a set of fixed strings.
        """
        # XXX: it would be nice if we could not do this remapping.
        return self.native.outcome or "none"

    @outcome.setter
    def outcome(self, new_value):
        """
        set outcome of the job to a new value
        """
        # XXX: it would be nice if we could not do this remapping.
        if new_value == "none":
            new_value = None
        self.native = self.native.get_builder(outcome=new_value).get_result()

    def _outcome_changed(self, old, new):
        """
        Internal method called when the value of self.native.outcome changes

        It sends the DBus PropertiesChanged signal for the 'outcome' property.
        """
        self.PropertiesChanged(JOB_RESULT_IFACE, {
            self.__class__.outcome._dbus_property: new
        }, [])

    @dbus.service.property(dbus_interface=JOB_RESULT_IFACE, signature="d")
    def execution_duration(self):
        """
        The amount of time in seconds it took to run this jobs command.

        :returns:
            The value of execution_duration or -1.0 if the command was not
            executed yet.
        """
        execution_duration = self.native.execution_duration
        if execution_duration is None:
            return -1.0
        else:
            return execution_duration

    @dbus.service.property(dbus_interface=JOB_RESULT_IFACE, signature="v")
    def return_code(self):
        """
        return code of the called program
        """
        value = self.native.return_code
        if value is None:
            return ""
        else:
            return value

    # comments are settable, useful thing that

    @dbus.service.property(dbus_interface=JOB_RESULT_IFACE, signature="s")
    def comments(self):
        """
        comment added by the operator
        """
        return self.native.comments or ""

    @comments.setter
    def comments(self, new_value):
        """
        set comments to a new value
        """
        self.native = self.native.get_builder(comments=new_value).get_result()

    def _comments_changed(self, old, new):
        """
        Internal method called when the value of self.native.comments changes

        It sends the DBus PropertiesChanged signal for the 'comments' property.
        """
        self.PropertiesChanged(JOB_RESULT_IFACE, {
            self.__class__.comments._dbus_property: new
        }, [])

    @dbus.service.property(
        dbus_interface=JOB_RESULT_IFACE, signature="a(dsay)")
    def io_log(self):
        """
        The input-output log.

        Contains a record of all of the output (actually,
        it has no input logs) that was sent by the called program.

        The format is: array<struct<double, string, array<bytes>>>
        """
        return dbus.types.Array(self.native.get_io_log(), signature="(dsay)")


class JobStateWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing JobState objects on DBus.

    Each job state wrapper has two related objects, job definition and job
    result. The job state wrapper itself is not a object manager so all of
    the managed objects belong in the session this job state is associated
    with.

    The job property is immutable. The result property is mutable but only
    through native python API. Any changes to the result property are
    propagated to changes in the DBus layer. When the 'result' property
    changes it is not reflecting changes of the object referenced by that
    property (those are separate) but instead indicates that the whole
    referenced object has been replaced by another object.

    Since JobStateWrapper is not an ObjectManager it does not manage the
    exact lifecycle and does not keep a collection that would reference
    the various result objects it must delegate this task to an instance
    of SessionWrapper (the instance that it is associated with).
    """

    HIDDEN_INTERFACES = frozenset([
        OBJECT_MANAGER_IFACE,
    ])

    def __shared_initialize__(self, session_wrapper, **kwargs):
        # We need a reference to the session wrapper so that we can
        # react to result changes.
        self._session_wrapper = session_wrapper
        # Let's cache (and hold) references to the wrappers that
        # we should know about. This keeps them in the live set and makes
        # accessing relevant properties faster.
        self._result_wrapper = self.find_wrapper_by_native(self.native.result)
        self._job_wrapper = self.find_wrapper_by_native(self.native.job)
        # Connect to the on_result_changed signal so that we can keep the
        # referenced 'result' wrapper in sync with the native result object.
        self.native.on_result_changed.connect(self._result_changed)

    def __del__(self):
        super(JobStateWrapper, self).__del__()
        self.native.on_result_changed.disconnect(self._result_changed)

    def publish_related_objects(self, connection):
        super(JobStateWrapper, self).publish_related_objects(connection)
        self._result_wrapper.publish_related_objects(connection)
        self._job_wrapper.publish_related_objects(connection)

    # Value added

    @dbus.service.method(
        dbus_interface=JOB_STATE_IFACE, in_signature='', out_signature='b')
    def CanStart(self):
        """
        Quickly check if the associated job can run right now.
        """
        return self.native.can_start()

    @dbus.service.method(
        dbus_interface=JOB_STATE_IFACE, in_signature='', out_signature='s')
    def GetReadinessDescription(self):
        """
        Get a human readable description of the current readiness state
        """
        return self.native.get_readiness_description()

    @dbus.service.property(dbus_interface=JOB_STATE_IFACE, signature='o')
    @PlainBoxObjectWrapper.translate
    def job(self) -> 'o':
        """
        Job associated with this state
        """
        return self.native.job

    @dbus.service.property(dbus_interface=JOB_STATE_IFACE, signature='o')
    @PlainBoxObjectWrapper.translate
    def result(self) -> 'o':
        """
        Result of running the associated job
        """
        return self.native.result

    def _result_changed(self, old, new):
        """
        Internal method called when the value of self.native.comments changes

        It ensures that we have appropriate wrapper for the new result wrapper
        and that it is properly accounted for by the session. It also sends
        the DBus PropertiesChanged signal for the 'result' property.
        """
        logger.debug("_result_changed(%r, %r)", old, new)
        # Add the new result object
        try:
            result_wrapper = self.find_wrapper_by_native(new)
        except KeyError:
            result_wrapper = self._session_wrapper.add_result(new)
        # Notify applications that the result property has changed
        self.PropertiesChanged(JOB_STATE_IFACE, {
            self.__class__.result._dbus_property: result_wrapper
        }, [])
        # Remove the old result object
        try:
            self._session_wrapper.remove_result(old)
        except KeyError:
            # NOTE: the result may have been self-removed on earlier
            # assignment to outcome. In that case this is a non-fatal
            # problem and we can just carry on.
            logger.warning("(_result_changed) Unable to remove old result: %r", old)
            pass

    @dbus.service.property(dbus_interface=JOB_STATE_IFACE, signature='a(isss)')
    def readiness_inhibitor_list(self):
        """
        The list of readiness inhibitors of the associated job

        The list is represented as an array of structures. Each structure
        has a integer and two strings. The integer encodes the cause
        of inhibition.

        Cause may have one of the following values:

        0 - UNDESIRED:
            This job was not selected to run in this session

        1 - PENDING_DEP:
           This job depends on another job which was not started yet

        2 - FAILED_DEP:
            This job depends on another job which was started and failed

        3 - PENDING_RESOURCE:
            This job has a resource requirement expression that uses a resource
            produced by another job which was not started yet

        4 - FAILED_RESOURCE:
            This job has a resource requirement that evaluated to a false value

        The next two strings are the name of the related job and the name
        of the related expression. Either may be empty.
        """
        return dbus.types.Array([
            (inhibitor.cause,
             inhibitor.cause.name,
             (inhibitor.related_job.id
              if inhibitor.related_job is not None else ""),
             (inhibitor.related_expression.text
              if inhibitor.related_expression is not None else ""))
            for inhibitor in self.native.readiness_inhibitor_list
        ], signature="(isss)")


class SessionWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing SessionState objects on DBus
    """

    HIDDEN_INTERFACES = frozenset()

    # XXX: those will change to SessionManager later and session state will be
    # a part of that (along with session storage)

    def __shared_initialize__(self, **kwargs):
        # Wrap the initial set of objects reachable via the session state map
        # We don't use the add_{job,result,state}() methods as they also
        # change managed_object_list and we just want to send one big event
        # rather than a storm of tiny events.
        self._job_state_map_wrapper = {}
        for job_name, job_state in self.native.job_state_map.items():
            # NOTE: we assume that each job is already wrapped by its provider
            job_wrapper = self.find_wrapper_by_native(job_state.job)
            assert job_wrapper is not None
            # Wrap the result and the state object
            result_wrapper = self._maybe_wrap(job_state.result)
            assert result_wrapper is not None
            state_wrapper = self._maybe_wrap(job_state)
            self._job_state_map_wrapper[job_name] = state_wrapper
        # Keep track of new jobs as they are added to the session
        self.native.on_job_added.connect(self._job_added)
        self.native.on_job_removed.connect(self._job_removed)

    def publish_related_objects(self, connection):
        super(SessionWrapper, self).publish_related_objects(connection)
        # Publish all the JobState wrappers and their related objects
        for job_state in self._job_state_map_wrapper.values():
            job_state.publish_related_objects(connection)

    def publish_managed_objects(self):
        wrapper_list = list(self._iter_wrappers())
        self.add_managed_object_list(wrapper_list)
        for wrapper in wrapper_list:
            wrapper.publish_managed_objects()

    def _iter_wrappers(self):
        return itertools.chain(
            # Get all of the JobState wrappers
            self._job_state_map_wrapper.values(),
            # And all the JobResult wrappers
            [self.find_wrapper_by_native(job_state_wrapper.native.result)
             for job_state_wrapper in self._job_state_map_wrapper.values()])

    def add_result(self, result):
        """
        Add a result representation to DBus.

        Take a IJobResult subclass instance, wrap it in JobResultWrapper,
        publish it so that it shows up on DBus, add it to the collection
        of objects managed by this SessionWrapper so that it sends
        InterfacesAdded signals and can be enumerated with
        GetManagedObjects() and return the wrapper to the caller.

        :returns:
            The wrapper for the result that was added
        """
        logger.info(_("Adding result %r to DBus"), result)
        result_wrapper = self._maybe_wrap(result)
        result_wrapper.publish_self(self.connection)
        self.add_managed_object(result_wrapper)
        return result_wrapper

    def remove_result(self, result):
        """
        Remove a result representation from DBus.

        Take a IJobResult subclass instance, find the JobResultWrapper that
        it is wrapped in. Remove it from the collection of objects managed
        by this SessionWrapper so that it sends InterfacesRemoved signal
        and can no longer be enumerated with GetManagedObjects(), remove it
        from the bus and return the wrapper to the caller.

        :returns:
            The wrapper for the result that was removed
        """
        logger.info(_("Removing result %r from DBus"), result)
        result_wrapper = self.find_wrapper_by_native(result)
        self.remove_managed_object(result_wrapper)
        result_wrapper.remove_from_connection()
        return result_wrapper

    def add_job(self, job):
        """
        Add a job representation to DBus.

        :param job:
            Job to add to the bus
        :ptype job:
            JobDefinition

        Take a JobDefinition, wrap it in JobResultWrapper, publish it so that
        it shows up on DBus, add it to the collection of objects managed by
        this SessionWrapper so that it sends InterfacesAdded signals and
        can be enumerated with GetManagedObjects.

        :returns:
            The wrapper for the job that was added
        """
        logger.info(_("Adding job %r to DBus"), job)
        job_wrapper = self._maybe_wrap(job)
        # Mark this job as generated, so far we only add generated jobs at
        # runtime and we need to treat those differently when we're changing
        # the session.
        job_wrapper._is_generated = True
        job_wrapper.publish_self(self.connection)
        self.add_managed_object(job_wrapper)
        return job_wrapper

    def add_state(self, state):
        """
        Add a job state representation to DBus.

        :param state:
            Job state to add to the bus
        :ptype state:
            JobState
        :returns:
            The wrapper for the job that was added

        Take a JobState, wrap it in JobStateWrapper, publish it so that it
        shows up on DBus, add it to the collection of objects managed by this
        SessionWrapper so that it sends InterfacesAdded signals and can be
        enumerated with GetManagedObjects.

        .. note::
            This method must be called after both result and job definition
            have been added (presumably with :meth:`add_job()`
            and :meth:`add_result()`). This method *does not* publish those
            objects, it only publishes the state object.
        """
        logger.info(_("Adding job state %r to DBus"), state)
        state_wrapper = self._maybe_wrap(state)
        state_wrapper.publish_self(self.connection)
        self.add_managed_object(state_wrapper)
        return state_wrapper

    def _maybe_wrap(self, obj):
        """
        Wrap a native object in the appropriate DBus wrapper.

        :param obj:
            The object to wrap
        :ptype obj:
            JobDefinition, IJobResult or JobState
        :returns:
            The wrapper associated with the object

        The object is only wrapped if it was not wrapped previously (at most
        one wrapper is created for any given native object). Only a few classes
        are supported, those are JobDefinition, IJobResult, JobState.
        """
        try:
            return self.find_wrapper_by_native(obj)
        except LookupError:
            if isinstance(obj, JobDefinition):
                return JobDefinitionWrapper(obj)
            elif isinstance(obj, IJobResult):
                return JobResultWrapper(obj)
            elif isinstance(obj, JobState):
                return JobStateWrapper(obj, session_wrapper=self)
            else:
                raise TypeError(
                    _("Unable to wrap object of type {0}").format(type(obj)))

    def _job_added(self, job):
        """
        Internal method connected to the SessionState.on_job_added() signal.

        This method is called when a generated job is added to the session.
        This method adds the corresponding job definition, job result and
        job state to the bus and sends appropriate notifications.
        """
        logger.debug("_job_added(%r)", job)
        # Get references to the three key objects, job, state and result
        state = self.native.job_state_map[job.id]
        result = state.result
        assert job is state.job
        # Wrap them in the right order (state has to be last)
        self.add_job(job)
        self.add_result(result)
        state_wrapper = self.add_state(state)
        # Update the job_state_map wrapper that we have here
        self._job_state_map_wrapper[job.id] = state_wrapper
        # Send the signal that the 'job_state_map' property has changed
        self.PropertiesChanged(SESSION_IFACE, {
            self.__class__.job_state_map._dbus_property:
            self._job_state_map_wrapper
        }, [])

    def _job_removed(self, job):
        """
        Internal method connected to the SessionState.on_job_removed() signal.

        This method is called (so far) only when the list of jobs is trimmed
        after doing calling :meth:`Resume()`. This method looks up the
        associated state and result object and removes them. If the removed job
        was not a part of the provider set (it was a generated job) it is also
        removed. Lastly this method sends the appropriate notifications.
        """
        logger.debug("_job_removed(%r)", job)
        # Get references to the three key objects, job, state and result
        state_wrapper = self._job_state_map_wrapper[job.id]
        result_wrapper = state_wrapper._result_wrapper
        job_wrapper = state_wrapper._job_wrapper
        # Remove result and state from our managed object list
        self.remove_managed_object(result_wrapper)
        self.remove_managed_object(state_wrapper)
        # Remove job from managed object list if it was generated
        if job_wrapper._is_generated:
            self.remove_managed_object(job_wrapper)
        # Remove result and state wrappers from dbus
        result_wrapper.remove_from_connection()
        state_wrapper.remove_from_connection()
        # Remove job from dbus if it was generated
        if job_wrapper._is_generated:
            job_wrapper.remove_from_connection()
        # Update the job_state_map wrapper that we have here
        del self._job_state_map_wrapper[job.id]
        # Send the signal that the 'job_state_map' property has changed
        self.PropertiesChanged(SESSION_IFACE, {
            self.__class__.job_state_map._dbus_property:
            self._job_state_map_wrapper
        }, [])

    # Value added

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='ao', out_signature='as')
    @PlainBoxObjectWrapper.translate
    def UpdateDesiredJobList(self, desired_job_list: 'ao'):
        logger.info("UpdateDesiredJobList(%r)", desired_job_list)
        problem_list = self.native.update_desired_job_list(desired_job_list)
        # TODO: map each problem into a structure (check which fields should be
        # presented). Document this in the docstring.
        return [str(problem) for problem in problem_list]

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='oo', out_signature='')
    @PlainBoxObjectWrapper.translate
    def UpdateJobResult(self, job: 'o', result: 'o'):
        logger.info("UpdateJobResult(%r, %r)", job, result)
        self.native.update_job_result(job, result)

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='', out_signature='(dd)')
    def GetEstimatedDuration(self):
        automated, manual = self.native.get_estimated_duration()
        if automated is None:
            automated = -1.0
        if manual is None:
            manual = -1.0
        return automated, manual

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='', out_signature='s')
    def PreviousSessionFile(self):
        # TODO: this method makes no sense here, it should not be on a session
        # object, it should, if anything, be on the service object.
        logger.info("PreviousSessionFile()")
        previous_session_file = self.native.previous_session_file()
        logger.info("PreviousSessionFile() -> %r", previous_session_file)
        if previous_session_file:
            return previous_session_file
        else:
            return ''

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='', out_signature='')
    def Resume(self):
        self.native.resume()

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='', out_signature='')
    def Clean(self):
        logger.info("Clean()")
        self.native.clean()

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='', out_signature='')
    def Remove(self):
        logger.info("Remove()")
        # Disconnect all signals listeners from the native session object
        remove_signals_listeners(self)
        for wrapper in self.managed_objects:
            wrapper.remove_from_connection()
        self.remove_from_connection()
        self.native.remove()
        logger.debug(_("Remove() completed"))

    @dbus.service.method(
        dbus_interface=SESSION_IFACE, in_signature='', out_signature='')
    def PersistentSave(self):
        logger.info("PersistentSave()")
        self.native.persistent_save()

    @dbus.service.property(dbus_interface=SESSION_IFACE, signature='ao')
    @PlainBoxObjectWrapper.translate
    def job_list(self) -> 'ao':
        return self.native.job_list

    # TODO: signal<run_list>

    @dbus.service.property(dbus_interface=SESSION_IFACE, signature='ao')
    @PlainBoxObjectWrapper.translate
    def desired_job_list(self) -> 'ao':
        return self.native.desired_job_list

    # TODO: signal<run_list>

    @dbus.service.property(dbus_interface=SESSION_IFACE, signature='ao')
    @PlainBoxObjectWrapper.translate
    def run_list(self) -> 'ao':
        return self.native.run_list

    # TODO: signal<run_list>

    @dbus.service.property(dbus_interface=SESSION_IFACE, signature='a{so}')
    @PlainBoxObjectWrapper.translate
    def job_state_map(self) -> 'a{so}':
        return self.native.job_state_map

    @dbus.service.property(dbus_interface=SESSION_IFACE, signature='a{sv}')
    def metadata(self):
        return dbus.types.Dictionary({
            'title': self.native.metadata.title or "",
            'flags': dbus.types.Array(
                sorted(self.native.metadata.flags), signature='s'),
            'running_job_name': self.native.metadata.running_job_name or "",
            'app_blob': self.native.metadata.app_blob or b'',
            'app_id': self.native.metadata.app_id or ''
        }, signature="sv")

    @metadata.setter
    def metadata(self, value):
        self.native.metadata.title = value['title']
        self.native.metadata.running_job_name = value['running_job_name']
        self.native.metadata.flags = value['flags']
        self.native.metadata.app_blob = bytes(value.get('app_blob', b''))
        self.native.metadata.app_id = value.get('app_id', '')

    # TODO: signal<metadata>

    @dbus.service.signal(
        dbus_interface=SESSION_IFACE, signature='os')
    def AskForOutcome(self, primed_job: 'o', suggested_outcome: 's'):
        """
        Signal sent when the user should be consulted for the outcome.

        The signal carries:
        - the primed_job instance (which is the sender of this signal anyway).
        - the suggested_outcome for the test based on the execution of the
          test command if it exists.

        This signal triggers important interactions in the GUI, the typical
        use case for the suggested_outcome is:

        - When the "test" button is clicked on a manual test, the outcome
          (e.g. the yes/no/skip radiobox) needs to be automatically updated to
          reflect actual test result. Otherwise, a failing test will not be
          detected by the user, which will cause embarrassment.
        """
        logger.info(
            _("AskForOutcome(%r) suggested outcome is (%s)"),
            primed_job, suggested_outcome)

    @dbus.service.signal(
        dbus_interface=SESSION_IFACE, signature='o')
    def ShowInteractiveUI(self, primed_job: 'o'):
        """
        Signal sent when the test requires user interaction.

        The signal carries:
        - the primed_job instance (which is the sender of this signal anyway).

        This signal triggers important interactions in the GUI.
        """
        logger.info("ShowInteractiveUI(%r)", primed_job)


class ProviderWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing Provider1 objects on DBus
    """

    HIDDEN_INTERFACES = frozenset()

    def __shared_initialize__(self, **kwargs):
        self._job_wrapper_list = [
            JobDefinitionWrapper(job) for job in self.native.job_list]
        self._whitelist_wrapper_list = [
            TestPlanWrapper(unit)
            for unit in self.native.unit_list if unit.Meta.name == 'test plan']

    def _get_preferred_object_path(self):
        mangled_name = mangle_object_path(self.native.name)
        return "/plainbox/provider/{}".format(mangled_name)

    def publish_related_objects(self, connection):
        super(ProviderWrapper, self).publish_related_objects(connection)
        wrapper_list = list(self._iter_wrappers())
        for wrapper in wrapper_list:
            wrapper.publish_related_objects(connection)

    def publish_managed_objects(self):
        wrapper_list = list(self._iter_wrappers())
        self.add_managed_object_list(wrapper_list)

    def _iter_wrappers(self):
        return itertools.chain(
            self._job_wrapper_list,
            self._whitelist_wrapper_list)

    # Value added

    @dbus.service.property(dbus_interface=PROVIDER_IFACE, signature="s")
    def name(self):
        """
        name of this provider
        """
        return self.native.name

    @dbus.service.property(dbus_interface=PROVIDER_IFACE, signature="s")
    def description(self):
        """
        description of this provider
        """
        return self.native.description

    @dbus.service.property(dbus_interface=PROVIDER_IFACE, signature="s")
    def gettext_domain(self):
        """
        the name of the gettext domain associated with this provider

        This value may be empty, in such case provider data cannot be localized
        for the user environment.
        """
        return self.native.gettext_domain or ""


class ServiceWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing Service objects on DBus
    """

    HIDDEN_INTERFACES = frozenset()

    # Internal setup stuff

    def __shared_initialize__(self, on_exit, **kwargs):
        self._on_exit = on_exit
        self._provider_wrapper_list = [
            ProviderWrapper(provider)
            for provider in self.native.provider_list]

    def _get_preferred_object_path(self):
        return "/plainbox/service1"

    def publish_related_objects(self, connection):
        super(ServiceWrapper, self).publish_related_objects(connection)
        for wrapper in self._provider_wrapper_list:
            wrapper.publish_related_objects(connection)

    def publish_managed_objects(self):
        # First publish all of our providers
        self.add_managed_object_list(self._provider_wrapper_list)
        # Then ask the providers to publish their own objects
        for wrapper in self._provider_wrapper_list:
            wrapper.publish_managed_objects()

    # Value added

    @dbus.service.property(dbus_interface=SERVICE_IFACE, signature="s")
    def version(self):
        """
        version of this provider
        """
        return self.native.version

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='', out_signature='')
    def Exit(self):
        """
        Shut down the service and terminate
        """
        # TODO: raise exception when job is in progress
        logger.info("Exit()")
        self.native.close()
        self._on_exit()

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='', out_signature='a{sas}')
    def GetAllExporters(self):
        """
        Get all exporters names and their respective options
        """
        return self.native.get_all_exporters()

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='osas', out_signature='s')
    @PlainBoxObjectWrapper.translate
    def ExportSession(self, session: 'o', output_format: 's',
                      option_list: 'as'):
        return self.native.export_session(session, output_format, option_list)

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='osasss', out_signature='s')
    @PlainBoxObjectWrapper.translate
    def ExportSessionToFile(self, session: 'o', output_format: 's',
                            option_list: 'as', exporter_unit: 's',
                            output_file: 's'):
        return self.native.export_session_to_file(
            session, output_format, option_list, exporter_unit, output_file)

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='', out_signature='as')
    def GetAllTransports(self):
        """
        Get all transports names and their respective options
        """
        return self.native.get_all_transports()

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='ossss', out_signature='s')
    @PlainBoxObjectWrapper.translate
    def SendDataViaTransport(self, session: 'o', transport: 's', where: 's',
                             options: 's', data: 's'):
        return self.native.send_data_via_transport(session, transport, where,
                                                   options, data)

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='ao', out_signature='o')
    @PlainBoxObjectWrapper.translate
    def CreateSession(self, job_list: 'ao'):
        # remove existing sessions from sessions_list
        if self.native.session_list:
            del self.native.session_list[:]
        logger.info("CreateSession(%r)", job_list)
        # Create a session
        session_obj = self.native.create_session(job_list)
        # Keep a global reference around so that job.via can be emulated
        self.__class__._session_obj = session_obj
        # Wrap it
        session_wrp = SessionWrapper(session_obj)
        # Publish all objects
        session_wrp.publish_related_objects(self.connection)
        # Announce the session is there
        self.add_managed_object(session_wrp)
        # Announce any session children
        session_wrp.publish_managed_objects()
        # Return the session wrapper back
        return session_wrp

    # Private class-wide instance of the session that gets created by
    # :meth:`CreateSession()`.
    _session_obj = None

    @classmethod
    def get_session_obj(cls):
        """
        Get the non-wrapped (native) session object associated with the most
        recently created instance of the service wrapper's
        :meth:`CreateSession()` method

        :returns:
            The Session object (or None if it's not created yet).
        """
        return cls._session_obj

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='oo', out_signature='o')
    @PlainBoxObjectWrapper.translate
    def PrimeJob(self, session: 'o', job: 'o') -> 'o':
        logger.info("PrimeJob(%r, %r)", session, job)
        # Get a primed job for the arguments we've got...
        primed_job = self.native.prime_job(session, job)
        # ...wrap it for DBus...
        primed_job_wrapper = PrimedJobWrapper(
            primed_job, session_wrapper=self.find_wrapper_by_native(session))
        # ...publish it...
        primed_job_wrapper.publish_self(self.connection)
        # Call the method that decides on what to really do, see the docstring
        # for details. This cannot be called inside __init__() as we need to
        # publish the wrapper first. When that happens this method can safely
        # send signals.
        primed_job_wrapper._decide_on_what_to_do()
        # ...and return it
        return primed_job

    RunJob = PrimeJob

    @dbus.service.method(
        dbus_interface=SERVICE_IFACE, in_signature='ao', out_signature='ao')
    @PlainBoxObjectWrapper.translate
    def SelectJobs(self, whitelist_list: 'ao') -> 'ao':
        """
        Compute the effective desired job list out of a list of (arbitrary)
        desired whitelists or job definitions.

        :param whitelist_list:
            A list of jobs or whitelists to select. Each whitelist selects all
            the jobs selected by that whitelist.

            This argument is a simple, limited, encoding of job qualifiers that
            is sufficient to implement the desired semantics of the Checkbox
            GUI.

        :returns:
            A list of jobs that were selected.
        """
        if len(self.native.session_list) > 0:
            legacy_session = self.native.session_list[0]
            job_list = legacy_session.job_list
            # NOTE: this is a bit magic so an explanation is in order.
            # Each time SelectJobs is called by checkbox-gui we use it
            # as a hack to teach the SessionManager hidden inside
            # the SessionStateLegacyAPICompatImpl object (which is just
            # a legacy SessionState class implemented on top of the new
            # API) about test plans that this application is using.
            #
            # This results in a low-cost way to have overrides working
            # in checkbox-gui without touching the C++ side of it.
            if legacy_session._manager is not None:
                legacy_session._manager.test_plans = tuple(whitelist_list)
        else:
            job_list = list(itertools.chain(*[
                p.job_list for p in self.native.provider_list]))
        return select_jobs(job_list,
                           [w.get_qualifier() for w in whitelist_list])


class UIOutputPrinter(extcmd.DelegateBase):
    """
    Delegate for extcmd that redirect all output to the UI.
    """

    def __init__(self, runner):
        self._lineno = collections.defaultdict(int)
        self._runner = runner

    def on_line(self, stream_name, line):
        # FIXME: this is not a line number,
        # TODO: tie this into existing code in runner.py (the module)
        self._lineno[stream_name] += 1
        self._runner.IOLogGenerated(self._lineno[stream_name],
                                    stream_name, line)


class PrimedJobWrapper(PlainBoxObjectWrapper):
    """
    Wrapper for exposing PrimedJob objects on DBus
    """

    HIDDEN_INTERFACES = frozenset(
        OBJECT_MANAGER_IFACE,
    )

    def __shared_initialize__(self, session_wrapper, **kwargs):
        # SessionWrapper instance, we're using it to publish/unpublish results
        # on DBus as they technically belong to the session.
        self._session_wrapper = session_wrapper
        # A result object we got from running the command OR the result this
        # job used to have before. It should be always published on the bus.
        self._result = None
        self._result_builder = None
        # A future for the result each time we're waiting for the command to
        # finish. Gets reset to None after the command is done executing.
        self._result_future = None

    @dbus.service.method(
        dbus_interface=RUNNING_JOB_IFACE, in_signature='', out_signature='')
    def Kill(self):
        """
        Unused method.

        Could be used to implement a way to cancel the command.
        """
        # NOTE: this should:
        # 1) attempt to cancel the future in the extremely rare case where it
        #    is not started yet
        # 2) kill the job otherwise
        logger.error(_("Kill() is not implemented"))

    @dbus.service.method(
        dbus_interface=RUNNING_JOB_IFACE, in_signature='', out_signature='')
    def RunCommand(self):
        """
        Method invoked by the GUI each time the "test" button is pressed.

        Two cases are possible here:

        1) The command isn't running. In this case we start the command and
           return. A callback will interpret the result once the command
           finishes executing (see :meth:`_legacy_result_juggle()`)

        2) The command is running. In that case we just ignore the call and log
           a warning. The GUI should not make the "test" button clickable while
           the command is runninig.
        """
        if self.native.job.automated:
            logger.error(
                _("RunCommand() should not be called for automated jobs"))
        if self._result_future is None:
            logger.info(_("RunCommand() is starting to run the job"))
            self._result_future = self._run_and_set_callback()
        else:
            logger.warning(
                _("RunCommand() ignored, waiting for command to finish"))

    # Legacy GUI behavior method.
    # Should be redesigned when we can change GUI internals
    @dbus.service.method(
        dbus_interface=RUNNING_JOB_IFACE, in_signature='ss', out_signature='')
    def SetOutcome(self, outcome, comments=None):
        """
        Method called by the GUI when the "continue" button is pressed.

        Three cases are possible:
        1) There is no result yet
        2) There is a result already
        3) The command is still running! (not handled)
        """
        logger.info("SetOutcome(%r, %r)", outcome, comments)
        if self._result_future is not None:
            logger.error(
                _("SetOutcome() called while the command is still"
                  " running!"))
        if self._result is None:
            # Warn us if this method is being called on jobs other than
            # 'manual' before we get the result after running RunCommand()
            if self.native.job.plugin != "manual":
                logger.warning(
                    _("SetOutcome() called before RunCommand()"))
                logger.warning(
                    # TRANSLATORS: don't translate 'manual' translate it as
                    # 'of type manual' and leave the 'manual' string
                    # intact.
                    _("But the job is not manual, it is %s"),
                    self.native.job.plugin)
            # Create a new result object
            self._result_builder = JobResultBuilder()
            self._result_builder.outcome = outcome
            self._result_builder.comments = comments
            self._result = self._result_builder.get_result()
            # Add the new result object to the bus
            self._session_wrapper.add_result(self._result)
        else:
            assert self._result_builder is not None
            # Set the values as requested
            self._result_builder.outcome = outcome
            self._result_builder.comments = comments
            self._result = self._result_builder.get_result()
            # Add the new result object to the bus
            self._session_wrapper.add_result(self._result)
        # Notify the application that the result is ready. This has to be
        # done unconditionally each time this method called.
        self.JobResultAvailable(
            self.find_wrapper_by_native(self.native.job),
            self.find_wrapper_by_native(self._result))

    # Legacy GUI behavior signal.
    # Should be redesigned when we can change GUI internals
    @dbus.service.property(dbus_interface=RUNNING_JOB_IFACE, signature="s")
    def outcome_from_command(self):
        """
        property that contains the 'outcome' of the result.
        """
        if self._result is not None:
            # TODO: make it so that we don't have to do this translation
            if self._result.outcome == IJobResult.OUTCOME_NONE:
                return "none"
            else:
                return self._result.outcome
        else:
            logger.warning(_("outcome_from_command() called too early!"))
            logger.warning(_("There is nothing to return yet"))
            return ""

    @dbus.service.signal(
        dbus_interface=SERVICE_IFACE, signature='dsay')
    def IOLogGenerated(self, delay, stream_name, data):
        """
        Signal sent when IOLogRecord is generated

        ..note::
            This is not called at all in this implementation.
        """
        logger.info("IOLogGenerated(%r, %r, %r)", delay, stream_name, data)

    # Legacy GUI behavior signal.
    # Should be redesigned when we can change GUI internals
    @dbus.service.signal(
        dbus_interface=SERVICE_IFACE, signature='oo')
    def JobResultAvailable(self, job: 'o', result: 'o'):
        """
        Signal sent when result of running a job is ready
        """
        logger.info("JobResultAvailable(%r, %r)", job, result)

    def _decide_on_what_to_do(self):
        """
        Internal method of PrimedJobWrapper.

        This methods decides on how to behave after we get initialized

        This is a bit of a SNAFU... :/

        The GUI runs jobs differently, depending on the 'plugin' value.
        It either expects to call Service.RunJob() and see the job running
        (so waiting for signals when it is done) or it calls Service.RunJob()
        *and* runs RunCommand() on the returned object (possibly many times).
        """
        # TODO: change this to depend on jobbox 'startup' property
        # http://jobbox.readthedocs.org/en/latest/jobspec.html#startup
        if self.native.job.startup_user_interaction_required:
            logger.info(
                _("Sending ShowInteractiveUI() and not starting the job..."))
            # Ask the application to show the interaction GUI
            self._session_wrapper.ShowInteractiveUI(self)
        else:
            logger.info(_("Running %r right away"), self.native.job)
            self._result_future = self._run_and_set_callback()

    def _run_and_set_callback(self):
        """
        Internal method of PrimedJobWrapper

        Starts the future for the job result, adds a callbacks to it and
        returns the future.
        """
        future = self.native.run()
        future.add_done_callback(self._result_ready)
        return future

    def _result_ready(self, result_future):
        """
        Internal method called when the result future becomes ready
        """
        logger.debug("_result_ready(%r)", result_future)
        if self._result is not None:
            # NOTE: I'm not sure how this would behave if someone were to
            # already assign the old result to any state objects.
            try:
                self._session_wrapper.remove_result(self._result)
            except KeyError:
                # NOTE: the result may have been self-removed on earlier
                # assignment to outcome. In that case this is a non-fatal
                # problem and we can just carry on.
                logger.warning("(_result_ready) Unable to remove old result: %r", self._result)
                pass
        # Unpack the result from the future
        self._result = result_future.result()
        self._result_builder = self._result.get_builder()
        # Add the new result object to the session wrapper (and to the bus)
        self._session_wrapper.add_result(self._result)
        # Reset the future so that RunCommand() can run the job again
        self._result_future = None
        # Now fiddle with the GUI notifications
        if self._result.outcome != IJobResult.OUTCOME_UNDECIDED:
            # NOTE: OUTCOME_UNDECIDED is never handled by this method as
            # the user should already see the manual test interaction
            # dialog on their screen. For all other cases we need to notify
            # the GUI that execution has finished and we are really just
            # done with testing.
            logger.debug(
                _("calling JobResultAvailable(%r, %r)"),
                self.native.job, self._result)
            self.JobResultAvailable(
                self.find_wrapper_by_native(self.native.job),
                self.find_wrapper_by_native(self._result))
        else:
            logger.debug(
                _("sending AskForOutcome() after job finished"
                  " running with OUTCOME_UNDECIDED"))
            # Convert the return of the command to the suggested_outcome
            # for the job
            if self._result.return_code == 0:
                suggested_outcome = IJobResult.OUTCOME_PASS
            else:
                suggested_outcome = IJobResult.OUTCOME_FAIL
            self._session_wrapper.AskForOutcome(self, suggested_outcome)



# This file is part of Checkbox.
#
# Copyright 2013-2014 Canonical Ltd.
# Written by:
#   Sylvain Pineau <sylvain.pineau@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.
#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox_ng.commands.cli` -- Command line sub-command
===========================================================

.. warning::

    THIS MODULE DOES NOT HAVE STABLE PUBLIC API
"""

from argparse import SUPPRESS
from gettext import gettext as _
from logging import getLogger
import argparse

from plainbox.impl.commands import PlainBoxCommand
from plainbox.impl.commands.cmd_checkbox import CheckBoxCommandMixIn
from plainbox.impl.commands.inv_check_config import CheckConfigInvocation

from checkbox_ng.commands.newcli import CliInvocation2


logger = getLogger("checkbox.ng.commands.cli")


class CliCommand(PlainBoxCommand, CheckBoxCommandMixIn):
    """
    Command for running tests using the command line UI.
    """
    gettext_domain = "checkbox-ng"

    def __init__(self, provider_loader, config_loader, settings):
        self.provider_loader = provider_loader
        self.config_loader = config_loader
        self.settings = settings

    def invoked(self, ns):
        # Run check-config, if requested
        if ns.check_config:
            retval = CheckConfigInvocation(self.config_loader).run()
            return retval
        return CliInvocation2(
            self.provider_loader, self.loader_config, ns, self.settings
        ).run()

    def register_parser(self, subparsers):
        parser = subparsers.add_parser(self.settings['subparser_name'],
                                       help=self.settings['subparser_help'])
        parser.set_defaults(command=self)
        parser.set_defaults(dry_run=False, new_ui=True)
        parser.add_argument(
            '--new-ui', help=argparse.SUPPRESS, action='store_true')
        parser.add_argument(
            '--old-ui', dest='new_ui', help=argparse.SUPPRESS,
            action='store_false')
        parser.add_argument(
            "--check-config",
            action="store_true",
            help=_("run check-config"))
        group = parser.add_argument_group(title=_("user interface options"))
        parser.set_defaults(color=None)
        group.add_argument(
            '--no-color', dest='color', action='store_false', help=SUPPRESS)
        group.add_argument(
            '--non-interactive', action='store_true',
            help=_("skip tests that require interactivity"))
        group.add_argument(
            '--dont-suppress-output', action="store_true", default=False,
            help=_("don't suppress the output of certain job plugin types"))
        # Call enhance_parser from CheckBoxCommandMixIn
        self.enhance_parser(parser)



# This file is part of Checkbox.
#
# Copyright 2013 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`plainbox.impl.dbus_ex.service` -- DBus Service Extensions
===============================================================
"""

import functools
import logging
import threading
import weakref

from plainbox.vendor import morris
import _dbus_bindings
import dbus
import dbus.exceptions
import dbus.service

from checkbox_ng.dbus_ex import INTROSPECTABLE_IFACE
from checkbox_ng.dbus_ex import OBJECT_MANAGER_IFACE
from checkbox_ng.dbus_ex import PROPERTIES_IFACE
# Note: use our own version of the decorators because
# vanilla versions choke on annotations
from checkbox_ng.dbus_ex.decorators import method, signal


# This is the good old standard python property decorator
_property = property

__all__ = [
    'Interface',
    'Object',
    'method',
    'property',
    'signal',
]

logger = logging.getLogger("checkbox.ng.dbus_ex")


class InterfaceType(dbus.service.InterfaceType):
    """
    Subclass of :class:`dbus.service.InterfaceType` that also handles
    properties.
    """

    def _reflect_on_property(cls, func):
        reflection_data = (
            '    <property name="{}" type="{}" access="{}"/>\n').format(
                func._dbus_property, func._signature,
                func.dbus_access_flag)
        return reflection_data


#Subclass of :class:`dbus.service.Interface` that also handles properties
Interface = InterfaceType('Interface', (dbus.service.Interface,), {})


class property:
    """
    property that handles DBus stuff
    """

    def __init__(self, signature, dbus_interface, dbus_property=None,
                 setter=False):
        """
        Initialize new dbus_property with the given interface name.

        If dbus_property is not specified it is set to the name of the
        decorated method. In special circumstances you may wish to specify
        alternate dbus property name explicitly.

        If setter is set to True then the implicit decorated function is a
        setter, not the default getter. This allows to define write-only
        properties.
        """
        self.__name__ = None
        self.__doc__ = None
        self._signature = signature
        self._dbus_interface = dbus_interface
        self._dbus_property = dbus_property
        self._getf = None
        self._setf = None
        self._implicit_setter = setter

    def __repr__(self):
        return "<dbus.service.property {!r}>".format(self.__name__)

    @_property
    def dbus_access_flag(self):
        """
        access flag of this DBus property

        :returns: either "readwrite", "read" or "write"
        :raises TypeError: if the property is ill-defined
        """
        if self._getf and self._setf:
            return "readwrite"
        elif self._getf:
            return "read"
        elif self._setf:
            return "write"
        else:
            raise TypeError(
                "property provides neither readable nor writable")

    @_property
    def dbus_interface(self):
        """
        name of the DBus interface of this DBus property
        """
        return self._dbus_interface

    @_property
    def dbus_property(self):
        """
        name of this DBus property
        """
        return self._dbus_property

    @_property
    def signature(self):
        """
        signature of this DBus property
        """
        return self._signature

    @_property
    def setter(self):
        """
        decorator for setter functions

        This property can be used to decorate additional method that
        will be used as a property setter. Otherwise properties cannot
        be assigned.
        """
        def decorator(func):
            self._setf = func
            return self
        return decorator

    @_property
    def getter(self):
        """
        decorator for getter functions

        This property can be used to decorate additional method that
        will be used as a property getter. It is only provider for parity
        as by default, the @dbus.service.property() decorator designates
        a getter function. This behavior can be controlled by passing
        setter=True to the property initializer.
        """
        def decorator(func):
            self._getf = func
            return self
        return decorator

    def __call__(self, func):
        """
        Decorate a getter function and return the property object

        This method sets __name__, __doc__ and _dbus_property.
        """
        self.__name__ = func.__name__
        if self.__doc__ is None:
            self.__doc__ = func.__doc__
        if self._dbus_property is None:
            self._dbus_property = func.__name__
        if self._implicit_setter:
            return self.setter(func)
        else:
            return self.getter(func)

    def __get__(self, instance, owner):
        if instance is None:
            return self
        else:
            if self._getf is None:
                raise dbus.exceptions.DBusException(
                    "property is not readable")
            return self._getf(instance)

    def __set__(self, instance, value):
        if self._setf is None:
            raise dbus.exceptions.DBusException(
                "property is not writable")
        self._setf(instance, value)

    # This little helper is here is to help :meth:`Object.Introspect()`
    # figure out how to handle properties.
    _dbus_is_property = True


def cache_in_class(func):
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        cls = self.__class__
        attr_name = '_cache_{}_{}'.format(cls.__name__, func.__name__)
        if not hasattr(cls, attr_name):
            retval = func(self, *args, **kwargs)
            setattr(cls, attr_name, retval)
        else:
            retval = getattr(cls, attr_name)
        return retval
    return wrapper


class Object(Interface, dbus.service.Object):
    """
    dbus.service.Object subclass that providers additional features.

    This class providers the following additional features:

    * Implementation of the PROPERTIES_IFACE. This includes the methods
      Get(), Set(), GetAll() and the signal PropertiesChanged()

    * Implementation of the OBJECT_MANAGER_IFACE. This includes the method
      GetManagedObjects() and signals InterfacesAdded() and
      InterfacesRemoved().

    * Tracking of object-path-to-object association using the new
      :meth:`find_object_by_path()` method

    * Selective activation of any of the above interfaces using
      :meth:`should_expose_interface()` method.

    * Improved version of the INTROSPECTABLE_IFACE that understands properties
    """

    def __init__(self, conn=None, object_path=None, bus_name=None):
        dbus.service.Object.__init__(self, conn, object_path, bus_name)
        self._managed_object_list = []

    # [ Public DBus methods of the INTROSPECTABLE_IFACE interface ]

    @method(
        dbus_interface=INTROSPECTABLE_IFACE,
        in_signature='', out_signature='s',
        path_keyword='object_path', connection_keyword='connection')
    @cache_in_class
    def Introspect(self, object_path, connection):
        """
        Return a string of XML encoding this object's supported interfaces,
        methods and signals.
        """
        logger.debug("Introspect(object_path=%r)", object_path)
        reflection_data = (
            _dbus_bindings.DBUS_INTROSPECT_1_0_XML_DOCTYPE_DECL_NODE)
        reflection_data += '<node name="%s">\n' % object_path
        interfaces = self._dct_entry
        for (name, funcs) in interfaces.items():
            # Allow classes to ignore certain interfaces This is useful because
            # this class implements all kinds of methods internally (for
            # simplicity) but does not really advertise them all directly
            # unless asked to.
            if not self.should_expose_interface(name):
                continue
            reflection_data += '  <interface name="%s">\n' % (name)
            for func in funcs.values():
                if getattr(func, '_dbus_is_method', False):
                    reflection_data += self.__class__._reflect_on_method(func)
                elif getattr(func, '_dbus_is_signal', False):
                    reflection_data += self.__class__._reflect_on_signal(func)
                elif getattr(func, '_dbus_is_property', False):
                    reflection_data += (
                        self.__class__._reflect_on_property(func))
            reflection_data += '  </interface>\n'
        for name in connection.list_exported_child_objects(object_path):
            reflection_data += '  <node name="%s"/>\n' % name
        reflection_data += '</node>\n'
        logger.debug("Introspect() returns: %s", reflection_data)
        return reflection_data

    # [ Public DBus methods of the PROPERTIES_IFACE interface ]

    @dbus.service.method(
        dbus_interface=dbus.PROPERTIES_IFACE,
        in_signature="ss", out_signature="v")
    def Get(self, interface_name, property_name):
        """
        Get the value of a property @property_name on interface
        @interface_name.
        """
        logger.debug(
            "%r.Get(%r, %r) -> ...",
            self, interface_name, property_name)
        try:
            props = self._dct_entry[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                dbus.PROPERTIES_IFACE,
                "No such interface {}".format(interface_name))
        try:
            prop = props[property_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                dbus.PROPERTIES_IFACE,
                "No such property {}:{}".format(
                    interface_name, property_name))
        try:
            value = prop.__get__(self, self.__class__)
        except dbus.exceptions.DBusException as exc:
            logger.error(
                "%r.Get(%r, %r) -> (exception) %r",
                self, interface_name, property_name, exc)
            raise
        except Exception as exc:
            logger.exception(
                "runaway exception from Get(%r, %r)",
                interface_name, property_name)
            raise dbus.exceptions.DBusException(
                dbus.PROPERTIES_IFACE,
                "Unable to get property interface/property {}:{}: {!r}".format(
                    interface_name, property_name, exc))
        else:
            logger.debug(
                "%r.Get(%r, %r) -> %r",
                self, interface_name, property_name, value)
            return value

    @dbus.service.method(
        dbus_interface=dbus.PROPERTIES_IFACE,
        in_signature="ssv", out_signature="")
    def Set(self, interface_name, property_name, value):
        logger.debug(
            "%r.Set(%r, %r, %r) -> ...",
            self, interface_name, property_name, value)
        try:
            props = self._dct_entry[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                dbus.PROPERTIES_IFACE,
                "No such interface {}".format(interface_name))
        try:
            # Map the real property name
            prop = {
                prop.dbus_property: prop
                for prop in props.values()
                if isinstance(prop, property)
            }[property_name]
            if not isinstance(prop, property):
                raise KeyError(property_name)
        except KeyError:
            raise dbus.exceptions.DBusException(
                dbus.PROPERTIES_IFACE,
                "No such property {}:{}".format(
                    interface_name, property_name))
        try:
            prop.__set__(self, value)
        except dbus.exceptions.DBusException:
            raise
        except Exception as exc:
            logger.exception(
                "runaway exception from %r.Set(%r, %r, %r)",
                self, interface_name, property_name, value)
            raise dbus.exceptions.DBusException(
                dbus.PROPERTIES_IFACE,
                "Unable to set property {}:{}: {!r}".format(
                    interface_name, property_name, exc))
        logger.debug(
            "%r.Set(%r, %r, %r) -> None",
            self, interface_name, property_name, value)

    @dbus.service.method(
        dbus_interface=dbus.PROPERTIES_IFACE,
        in_signature="s", out_signature="a{sv}")
    def GetAll(self, interface_name):
        logger.debug("%r.GetAll(%r)", self, interface_name)
        try:
            props = self._dct_entry[interface_name]
        except KeyError:
            raise dbus.exceptions.DBusException(
                dbus.PROPERTIES_IFACE,
                "No such interface {}".format(interface_name))
        result = {}
        for prop in props.values():
            if not isinstance(prop, property):
                continue
            prop_name = prop.dbus_property
            try:
                prop_value = prop.__get__(self, self.__class__)
            except:
                logger.exception(
                    "Unable to read property %r from %r", prop, self)
            else:
                result[prop_name] = prop_value
        return result

    @dbus.service.signal(
        dbus_interface=dbus.PROPERTIES_IFACE,
        signature='sa{sv}as')
    def PropertiesChanged(
            self, interface_name, changed_properties, invalidated_properties):
        logger.debug(
            "PropertiesChanged(%r, %r, %r)",
            interface_name, changed_properties, invalidated_properties)

    # [ Public DBus methods of the OBJECT_MANAGER_IFACE interface ]

    @dbus.service.method(
        dbus_interface=OBJECT_MANAGER_IFACE,
        in_signature="", out_signature="a{oa{sa{sv}}}")
    def GetManagedObjects(self):
        logger.debug("%r.GetManagedObjects() -> ...", self)
        result = {}
        for obj in self._managed_object_list:
            logger.debug("Looking for stuff exported by %r", obj)
            result[obj] = {}
            for iface_name in obj._dct_entry.keys():
                props = obj.GetAll(iface_name)
                if len(props):
                    result[obj][iface_name] = props
        logger.debug("%r.GetManagedObjects() -> %r", self, result)
        return result

    @dbus.service.signal(
        dbus_interface=OBJECT_MANAGER_IFACE,
        signature='oa{sa{sv}}')
    def InterfacesAdded(self, object_path, interfaces_and_properties):
        logger.debug("%r.InterfacesAdded(%r, %r)",
                     self, object_path, interfaces_and_properties)

    @dbus.service.signal(
        dbus_interface=OBJECT_MANAGER_IFACE, signature='oas')
    def InterfacesRemoved(self, object_path, interfaces):
        logger.debug("%r.InterfacesRemoved(%r, %r)",
                     self, object_path, interfaces)

    # [ Overridden methods of dbus.service.Object ]

    def add_to_connection(self, connection, path):
        """
        Version of dbus.service.Object.add_to_connection() that keeps track of
        all object paths.
        """
        with self._object_path_map_lock:
            # Super-call add_to_connection(). This can fail which is
            # okay as we haven't really modified anything yet.
            super(Object, self).add_to_connection(connection, path)
            # Touch self.connection, this will fail if the call above failed
            # and self._connection (mind the leading underscore) is still None.
            # It will also fail if the object is being exposed on multiple
            # connections (so self._connection is _MANY). We are interested in
            # the second check as _MANY connections are not supported here.
            self.connection
            # If everything is okay, just add the specified path to the
            # _object_path_to_object_map.
            self._object_path_to_object_map[path] = self

    def remove_from_connection(self, connection=None, path=None):
        with self._object_path_map_lock:
            # Touch self.connection, this triggers a number of interesting
            # checks, in particular checks for self._connection (mind the
            # leading underscore) being _MANY or being None. Both of those
            # throw an AttributeError that we can simply propagate at this
            # point.
            self.connection
            # Create a copy of locations. This is required because locations
            # are modified by remove_from_connection() which can also fail.  If
            # we were to use self.locations here directly we would have to undo
            # any changes if remove_from_connection() raises an exception.
            # Instead it is easier to first super-call remove_from_connection()
            # and then do what we need to at this layer, after
            # remove_from_connection() finishes successfully.
            locations_copy = list(self.locations)
            # Super-call remove_from_connection()
            super(Object, self).remove_from_connection(connection, path)
            # If either path or connection are none then treat them like
            # match-any wild-cards. The same logic is implemented in the
            # superclass version of this method.
            if path is None or connection is None:
                # Location is a tuple of at least two elements, connection and
                # path. There may be other elements added later so let's not
                # assume this is a simple pair.
                for location in locations_copy:
                    location_conn = location[0]
                    location_path = location[1]
                    # If (connection matches or is None)
                    # and (path matches or is None)
                    # then remove that association
                    if ((location_conn == connection or connection is None)
                            and (path == location_path or path is None)):
                        del self._object_path_to_object_map[location_path]
            else:
                # If connection and path were specified, just remove the
                # association from the specified path.
                del self._object_path_to_object_map[path]

    # [ Custom Extension Methods ]

    def should_expose_interface(self, iface_name):
        """
        Check if the specified interface should be exposed.

        This method controls which of the interfaces are visible as implemented
        by this Object. By default objects don't implement any interface expect
        for PEER_IFACE. There are two more interfaces that are implemented
        internally but need to be explicitly exposed: the PROPERTIES_IFACE and
        OBJECT_MANAGER_IFACE.

        Typically subclasses should NOT override this method, instead
        subclasses should define class-scope HIDDEN_INTERFACES as a
        frozenset() of classes to hide and remove one of the entries found in
        _STD_INTERFACES from it to effectively enable that interface.
        """
        return iface_name not in self.HIDDEN_INTERFACES

    @classmethod
    def find_object_by_path(cls, object_path):
        """
        Find and return the object that is exposed as object_path on any
        connection. Using multiple connections is not supported at this time.

        .. note::
            This obviously only works for objects exposed from the same
            application. The main use case is to have a way to lookup object
            paths that may be passed as arguments and also originate in the
            same application.
        """
        # XXX: ideally this would be per-connection method.
        with cls._object_path_map_lock:
            return cls._object_path_to_object_map[object_path]

    @_property
    def managed_objects(self):
        """
        list of of managed objects.

        This collection is a part of the OBJECT_MANAGER_IFACE. While it can be
        manipulated directly (technically) it should only be manipulated using
        :meth:`add_managed_object()`, :meth:`add_manage_object_list()`,
        :meth:`remove_managed_object()` and
        :meth:`remove_managed_object_list()` as they send appropriate DBus
        signals.
        """
        return self._managed_object_list

    def add_managed_object(self, obj):
        self.add_managed_object_list([obj])

    def remove_managed_object(self, obj):
        self.remove_managed_object_list([obj])

    def add_managed_object_list(self, obj_list):
        logger.debug("Adding managed objects: %s", obj_list)
        for obj in obj_list:
            if not isinstance(obj, Object):
                raise TypeError("obj must be of type {!r}".format(Object))
        old = self._managed_object_list
        new = list(old)
        new.extend(obj_list)
        self._managed_object_list = new
        self._on_managed_objects_changed(old, new)

    def remove_managed_object_list(self, obj_list):
        logger.debug("Removing managed objects: %s", obj_list)
        for obj in obj_list:
            if not isinstance(obj, Object):
                raise TypeError("obj must be of type {!r}".format(Object))
        old = self._managed_object_list
        new = list(old)
        for obj in obj_list:
            new.remove(obj)
        self._managed_object_list = new
        self._on_managed_objects_changed(old, new)

    # [ Custom Private Implementation Data ]

    _STD_INTERFACES = frozenset([
        INTROSPECTABLE_IFACE,
        OBJECT_MANAGER_IFACE,
        # TODO: peer interface is not implemented in this class
        # PEER_IFACE,
        PROPERTIES_IFACE
    ])

    HIDDEN_INTERFACES = frozenset([
        OBJECT_MANAGER_IFACE,
        PROPERTIES_IFACE
    ])

    # Lock protecting access to _object_path_to_object_map.
    # XXX: ideally this would be a per-connection attribute
    _object_path_map_lock = threading.Lock()

    # Map of object_path -> dbus.service.Object instances
    # XXX: ideally this would be a per-connection attribute
    _object_path_to_object_map = weakref.WeakValueDictionary()

    # [ Custom Private Implementation Methods ]

    @_property
    def _dct_key(self):
        """
        the key indexing this Object in Object.__class__._dbus_class_table
        """
        return self.__class__.__module__ + '.' + self.__class__.__name__

    @_property
    def _dct_entry(self):
        """
        same as self.__class__._dbus_class_table[self._dct_key]
        """
        return self.__class__._dbus_class_table[self._dct_key]

    @morris.signal
    def _on_managed_objects_changed(self, old_objs, new_objs):
        logger.debug("%r._on_managed_objects_changed(%r, %r)",
                     self, old_objs, new_objs)
        for obj in frozenset(new_objs) - frozenset(old_objs):
            ifaces_and_props = {}
            for iface_name in obj._dct_entry.keys():
                try:
                    props = obj.GetAll(iface_name)
                except dbus.exceptions.DBusException as exc:
                    logger.warning("Caught %r", exc)
                else:
                    ifaces_and_props[iface_name] = props
            self.InterfacesAdded(obj.__dbus_object_path__, ifaces_and_props)
        for obj in frozenset(old_objs) - frozenset(new_objs):
            ifaces = list(obj._dct_entry.keys())
            self.InterfacesRemoved(obj.__dbus_object_path__, ifaces)


class ObjectWrapper(Object):
    """
    Wrapper for a single python object which makes it easier to expose over
    DBus as a service. The object should be injected into something that
    extends dbus.service.Object class.

    The class maintains an association between each wrapper and native object
    and offers methods for converting between the two.
    """

    # Lock protecting access to _native_id_to_wrapper_map
    _native_id_map_lock = threading.Lock()

    # Man of id(wrapper.native) -> wrapper
    _native_id_to_wrapper_map = weakref.WeakValueDictionary()

    def __init__(self, native, conn=None, object_path=None, bus_name=None):
        """
        Create a new wrapper for the specified native object
        """
        super(ObjectWrapper, self).__init__(conn, object_path, bus_name)
        with self._native_id_map_lock:
            self._native_id_to_wrapper_map[id(native)] = self
        self._native = native

    @_property
    def native(self):
        """
        native python object being wrapped by this wrapper
        """
        return self._native

    @native.setter
    def native(self, new_native):
        with self._native_id_map_lock:
            if new_native is self._native:
                return
            del self._native_id_to_wrapper_map[id(self._native)]
            self._native = new_native
            self._native_id_to_wrapper_map[id(new_native)] = self

    @classmethod
    def find_wrapper_by_native(cls, native):
        """
        Find the wrapper associated with the specified native object
        """
        with cls._native_id_map_lock:
            return cls._native_id_to_wrapper_map[id(native)]



# This file is part of Checkbox.
#
# Copyright 2014 Canonical Ltd.
# Written by:
#   Zygmunt Krynicki <zygmunt.krynicki@canonical.com>
#
# Checkbox is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3,
# as published by the Free Software Foundation.

#
# Checkbox is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Checkbox.  If not, see <http://www.gnu.org/licenses/>.

"""
:mod:`checkbox_ng.launcher` -- launcher definition
==================================================
"""

from gettext import gettext as _
import logging

from checkbox_ng.config import SECURE_ID_PATTERN
from plainbox.impl.secure import config
from plainbox.impl.transport import get_all_transports


logger = logging.getLogger("checkbox.ng.launcher")


class LauncherDefinition(config.Config):
    """
    Launcher definition.

    Launchers are small executables using one of the available user interfaces
    as the interpreter. This class contains all the available options that can
    be set inside the launcher, that will affect the user interface at runtime.
    """

    title = config.Variable(
        section="welcome",
        help_text=_("Application Title"))

    text = config.Variable(
        section="welcome",
        help_text=_("Welcome Message"))

    whitelist_filter = config.Variable(
        section="suite",
        # TODO: valid regexp text validator
        help_text=_("Pattern that whitelists need to match to be displayed"))

    whitelist_selection = config.Variable(
        section="suite",
        # TODO: valid regexp text validator
        help_text=_("Pattern that whitelists need to match to be selected"))

    skip_whitelist_selection = config.Variable(
        section="suite",
        kind=bool,
        default=False,
        help_text=_("If enabled then suite selection screen is not displayed"))

    skip_test_selection = config.Variable(
        section="suite",
        kind=bool,
        default=False,
        help_text=_("If enabled then test selection screen is not displayed"))

    input_type = config.Variable(
        section="submission",
        # TODO: probably a choice validator
        help_text=_("Type of the input field?"))

    ok_btn_text = config.Variable(
        section="submission",
        help_text=_("Label on the 'send' button"))

    submit_to_hexr = config.Variable(
        section="submission",
        kind=bool,
        # TODO: default?
        help_text=_("If enabled then test results will be also sent to HEXR"))

    submit_to = config.Variable(
        section="transport",
        validator_list=[config.ChoiceValidator(get_all_transports().keys())],
        help_text=_("Where to submit the test results to"))

    # TODO: Add a validator to ensure it looks like a valid URL
    submit_url = config.Variable(
        section="transport",
        help_text=_("HTTP endpoint to submit data to, using the"
                    " transport specified with submit_to."))

    secure_id = config.Variable(
        section="submission",
        validator_list=[config.PatternValidator(SECURE_ID_PATTERN)],
        help_text=_("Secure ID to identify the system this"
                    " submission belongs to."))

    config_filename = config.Variable(
        section="config",
        help_text=_("Name of custom configuration file"))

    dont_suppress_output = config.Variable(
        section="ui", kind=bool, default=False,
        help_text=_("Don't suppress the output of certain job plugin types."))

    exporter = config.Section(
        help_text=_("Section with only exported unit ids as keys (no values)"))
